unit NSys;

{$IMPORTEDDATA OFF}
{$MINENUMSIZE 4}

interface

uses Windows, NTypes, NLib;

type
  EOSError = class(Exception)
    ErrNo: Integer;
    Path: WideString;
    constructor Create(ErrNo: Integer); overload;
    constructor Create(const Path: WideString = ''); overload;
  end;

  EFileNotFound = class(EOSError);

  EPipeError = class(EOSError);


  TPid = DWORD;

  TTime = record
    Year: Integer;
    Mon: Integer;
    MDay: Integer;
    Hour: Integer;
    Min: Integer;
    Sec: Integer;
    WDay: Integer;
    UTCOffset: Integer;
  end;

  TFileMode = set of (fmRead, fmWrite);

  TStdIO = (stdInput, stdOutput, stdError);

  TFile = class(TStream)
  protected
    FHandle: THandle;
    FName: WideString;
    IsStdHandle: Boolean;
    procedure CreateHandle(const Name: WideString; Mode: TFileMode; Create_: DWORD);
  public
    constructor Create(const Name: WideString; Mode: TFileMode=[fmWrite]);
    constructor Open(const Name: WideString; Mode: TFileMode=[fmRead]); overload;
    constructor Open(StdIO: TStdIO); overload;
    constructor OpenConOut;
    constructor OpenConIn;
    constructor CreateByHandle(Handle: THandle);
    destructor Destroy; override;
    function Read(var Buf; NBytes: Cardinal; var NRead: Cardinal): LongBool; override;
    function Write(const Buf; NBytes: Cardinal): Cardinal; override;
    function Seek(Offset: Int64; Whence: TSeekWhence=swSet): Int64; override;
    function Seekable: Boolean; override;
    function Size: Int64;
    function IsConsole: Boolean;
    function IsEchoEnabled: Boolean;
    procedure EnableEcho(Flag: Boolean);
    property Handle: THandle read FHandle;
    property Name: WideString read FName;
  end;

  TDirMode = set of (
    dmDirectory, dmReadByOwner, dmWriteByOwner, dmSymLink, dmJunction
  );

  TDir = record
    Name: WideString;
    Size: Int64;
    ATime: Int64;
    MTime: Int64;
    Mode: TDirMode;
  end;

  TListDir = class
  protected
    Handle: THandle;
    FindData: TWin32FindDataW;
    First: Boolean;
  public
    Path: WideString;
    constructor Create(const Path: WideString);
    destructor Destroy; override;
    function Each(var Dir: TDir): Boolean;
  end;

  TFileAccessRights = set of (farRead, farWrite, farExecute);

  TProcess = class
  protected
    ProcessInformation: TProcessInformation;
  public
    StdOutReader: TFile;
    // if use StdErr instead, may be confused with NSys.StdErr (compiler bug?)
    StdErrReader: TFile;
    constructor Create(const Args: array of WideString;
        const StdIn: TFile; const StdOut: TFile; const StdErr: TFile;
        ErrToOut: Boolean=False);
    destructor Destroy; override;
    function Wait: Integer;
    procedure Times(var System, User: Int64);
  end;

const
  LineSep = #13#10;
  DevNull = 'NUL';

var
  Argv: TWideStringArray;
  ProgName: WideString;
  StdIn, StdOut, StdErr: TFile;


function _CreateSystemException(P: PExceptionRecord): Exception;

function ProgPath: WideString;

procedure KillProcess(Pid: TPid);
procedure IgnoreKeyboardInterrupt;

procedure Remove(const Path: WideString);
procedure ReplaceFile(const Src: WideString; const Dst: WideString);
procedure CreateDir(const Path: WideString);
procedure ChDir(const Path: WideString);
function GetWd: WideString;
procedure CreatePipe(var Reader, Writer: TFile);

function SysError(Code: Integer): WideString;

procedure SetEnv(const Name: WideString; const Value: WideString);
function GetEnv(const Name: WideString): WideString;
function GetEnvs: TWideStringArray;

function Exists(const Path: WideString): Boolean;
function IsDir(const Path: WideString): Boolean;
function IsFile(const Path: WideString): Boolean;
function DirStat(const Path: WideString): TDir;
procedure DirWStat(const F: TFile; const Dir: TDir);
procedure ClearDir(var Dir: TDir);
function SameFile(const Path1: WideString; const Path2: WideString): Boolean;
function ReadLink(const Path: WideString): WideString;
function GetFileAccessRights(const Path: WideString): TFileAccessRights;

function Time: Int64;
function NanoSec: Int64;
procedure SetNanoSec(const NanoSec: Int64);
function UTCTime(const T: Int64): TTime;
function LocalTime(const T: Int64): TTime;
procedure Sleep(MilliSecs: Cardinal);

procedure SplitDrive(const Path: WideString; var Drive, Tail: WideString);
procedure SplitExt(const Path: WideString; var Root, Ext: WideString);
procedure SplitPath(const Path: WideString; var Head, Tail: WideString);
function JoinPath(const Paths: array of WideString): WideString;
function AbsPath(const Path: WideString): WideString;
function IsAbsPath(const Path: WideString): Boolean;

function StrColl(const S1, S2: string): Integer;
function StrCollW(const S1, S2: WideString): Integer;
function StrUpper(const S: string): string;
function StrUpperW(const S: WideString): WideString;

function DecodeUTF8(const S: string): WideString;
function EncodeUTF8(const S: WideString): string;

procedure Print(const S: string = ''; const End_: string = LineSep); overload;
procedure Print(const F: TStream; const S: string = '';
      const End_: string = LineSep); overload;
procedure Print(const S: string; const Args: array of const;
    const End_: string = LineSep); overload;
procedure Print(const F: TStream; const S: string; const Args: array of const;
    const End_: string = LineSep); overload;

function GetPass(const Prompt: string): string;

function GenRandom(NBytes: Cardinal): string;

implementation

uses WinAdd;

const
  PathSep = WideChar('\');
  AltPathSep = WideChar('/');
  DriveSep = WideChar(':');
  ExtSep = WideChar('.');

  UnixTimeDiff: Int64 = 116444736000000000;

constructor EOSError.Create(ErrNo: Integer);
begin
  Self.ErrNo := ErrNo;
  Msg := SysError(ErrNo);
end;

constructor EOSError.Create(const Path: WideString = '');
begin
  ErrNo := GetLastError;
  Msg := SysError(ErrNo);
  Self.Path := Path;
end;

{ private functions }

function _CreateSystemException(P: PExceptionRecord): Exception;
const
  msg: PWideChar = 'Exception %1!x! at %4!x!';
var
  buf: PWideChar;
  len: DWORD;
begin
  Result := Exception.Create;
  len := FormatMessageW(FORMAT_MESSAGE_FROM_STRING or
      FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_ARGUMENT_ARRAY,
      msg, 0, 0, @buf, 0, P);
  SetString(Result.Msg, buf, len);
  LocalFree(HLOCAL(buf));
end;

function FindLastPathSep(const Path: WideString): Integer;
begin
  Result := Length(Path);
  while (Result > 0) and not (Path[Result] in [PathSep, AltPathSep]) do
    Dec(Result);
end;

function FindLastSep(const Path: WideString; Sep: WideChar): Integer;
begin
  Result := Length(Path);
  while (Result > 0) and not (Path[Result] = Sep) do
    Dec(Result);
end;

function FileTimeToUnixTime(const T: TFileTime): Int64;
begin
  Result := (Int64(T) - UnixTimeDiff) div 10000000;
end;

function UnixTimeToFileTime(const T: Int64): TFileTime;
begin
  Int64(Result) := T * 10000000 + UnixTimeDiff;
end;

function FindDataToDir(const FindData: TWin32FindDataW): TDir;
begin
  with FindData do begin
    Result.Name := cFileName;
    ULARGE_INTEGER(Result.Size).LowPart := nFileSizeLow;
    ULARGE_INTEGER(Result.Size).HighPart := nFileSizeHigh;
    Result.Mode := [dmReadByOwner, dmWriteByOwner];
    if (dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0 then
      Include(Result.Mode, dmDirectory);
    if (dwFileAttributes and FILE_ATTRIBUTE_READONLY) <> 0 then
      Exclude(Result.Mode, dmWriteByOwner);
    if (dwFileAttributes and FILE_ATTRIBUTE_REPARSE_POINT) <> 0 then
      case dwReserved0 of
        IO_REPARSE_TAG_MOUNT_POINT: Include(Result.Mode, dmJunction);
        IO_REPARSE_TAG_SYMLINK: Include(Result.Mode, dmSymLink);
      end;
    Result.MTime := FileTimeToUnixTime(ftLastWriteTime);
    Result.ATime := FileTimeToUnixTime(ftLastAccessTime);
  end;
end;

{ public functions }

function ProgPath: WideString;
begin
  SetString(Result, nil, MAX_PATH);
  SetLength(Result, GetModuleFileNameW(0, PWideChar(Result), Length(Result)));
end;

procedure KillProcess(Pid: TPid);
var
  h: THandle;
begin
  h := OpenProcess(PROCESS_TERMINATE, False, Pid);
  if h = 0 then
    raise EOSError.Create;
  try
    if not TerminateProcess(h, 0) then
      raise EOSError.Create;
  finally
    CloseHandle(h);
  end;
end;

function _ConsoleCtrlHandler(CtrlType: DWORD): BOOL; stdcall;
begin
  Result := CtrlType = CTRL_C_EVENT;
end;

procedure IgnoreKeyboardInterrupt;
begin
  // use handler instead of nil to not inherited by child process
  SetConsoleCtrlHandler(@_ConsoleCtrlHandler, True);
end;

procedure Remove(const Path: WideString);
begin
  if IsDir(Path) then begin
    if not RemoveDirectoryW(PWideChar(Path)) then
      raise EOSError.Create(Path);
  end else
    if not DeleteFileW(PWideChar(Path)) then
      raise EOSError.Create(Path);
end;

procedure ReplaceFile(const Src: WideString; const Dst: WideString);
begin
  if not MoveFileExW(PWideChar(Src), PWideChar(Dst),
      MOVEFILE_REPLACE_EXISTING) then
    raise EOSError.Create;
end;

procedure CreateDir(const Path: WideString);
begin
  if not CreateDirectoryW(PWideChar(Path), nil) then
    raise EOSError.Create(Path);
end;

procedure ChDir(const Path: WideString);
begin
  if not SetCurrentDirectoryW(PWideChar(Path)) then
    raise EOSError.Create(Path);
end;

function GetWd: WideString;
begin
  SetString(Result, nil, GetCurrentDirectoryW(0, nil));
  if GetCurrentDirectoryW(Length(Result), PWideChar(Result)) = 0 then
    raise EOSError.Create;
  SetLength(Result, Length(Result)-1); // -1 for terminating null character
end;

procedure CreatePipe(var Reader, Writer: TFile);
var
  r, w :THandle;
begin
  if not Windows.CreatePipe(r, w, nil, 0) then
    raise EOSError.Create;
  Reader := TFile.CreateByHandle(r);
  Writer := TFile.CreateByHandle(w);
end;

function SysError(Code: Integer): WideString;
const
  unknown: PWideChar = 'Unknown error %1!d!';
var
  buf: PWideChar;
  len: DWORD;
begin
  len := FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM or
      FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_IGNORE_INSERTS or
      FORMAT_MESSAGE_MAX_WIDTH_MASK, nil, Code, 0, @buf, 0, nil);
  if len = 0 then
    len := FormatMessageW(FORMAT_MESSAGE_FROM_STRING or
        FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_ARGUMENT_ARRAY,
        unknown, 0, 0, @buf, 0, @Code);
  SetString(Result, buf, len);
  LocalFree(HLOCAL(buf));
end;

procedure SetEnv(const Name: WideString; const Value: WideString);
begin
  if not SetEnvironmentVariableW(PWideChar(Name), PWideChar(Value)) then
    raise EOSError.Create;
end;

function GetEnv(const Name: WideString): Widestring;
var
  len: DWORD;
begin
  len := GetEnvironmentVariableW(PWideChar(Name), nil, 0);
  if len = 0 then
    case GetLastError of
      ERROR_SUCCESS, ERROR_ENVVAR_NOT_FOUND:
        begin
          Result := '';
          Exit;
        end
    else
      raise EOSError.Create;
    end;
  SetString(Result, nil, len);
  len:= GetEnvironmentVariableW(PWideChar(Name), PWideChar(Result), len);
  if len = 0 then
    case GetLastError of
      ERROR_SUCCESS, ERROR_ENVVAR_NOT_FOUND:
        begin
          Result := '';
          Exit;
        end
    else
      raise EOSError.Create;
    end;
  SetLength(Result, len);
end;

function GetEnvs: TWideStringArray;
var
  i: Integer;
  envs, sp, p, q: PWideChar;
begin
  envs := GetEnvironmentStringsW;
  if envs = nil then
    raise EOSError.Create;

  SetLength(Result, 256);
  i := 0;
  p := envs;
  while p^ <> #0 do begin
    sp := p;
    while p^ <> #0 do
      Inc(p);

    q := sp;
    Inc(q); // skip begninng '='
    while q^ <> '=' do
      Inc(q);

    if i = Length(Result) then
      SetLength(Result, i + i div 4);
    SetString(Result[i], sp, q - sp);
    Inc(q);
    SetString(Result[i+1], q, p - q);
    Inc(i, 2);
    Inc(p);
  end;
  SetLength(Result, i);

  if not FreeEnvironmentStringsW(envs) then
    raise EOSError.Create;
end;

function Exists(const Path: WideString): Boolean;
begin
  Result := GetFileAttributesW(PWideChar(Path)) <> $FFFFFFFF;
end;

function IsDir(const Path: WideString): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributesW(PWideChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) <> 0);
end;

function IsFile(const Path: WideString): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributesW(PWideChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) = 0);
end;

function DirStat(const Path: WideString): TDir;
var
  finddata: TWin32FindDataW;
  h: THandle;
begin
  // check path exsistence since FindFirstFile accepts wildcard characters
  if GetFileAttributesW(PWideChar(Path)) <> $FFFFFFFF then begin
    h := FindFirstFileW(PWideChar(Path), finddata);
    if h <> INVALID_HANDLE_VALUE then begin
      FindClose(h);
      Result := FindDataToDir(finddata);
      Exit;
    end;
  end;
  raise EOSError.Create(Path);
end;

procedure DirWStat(const F: TFile; const Dir: TDir);
var
  ft: TFileTime;
begin
  // TODO: support atime
  with Dir do begin
    if MTime <> 0 then begin
      ft := UnixTimeToFileTime(MTime);
      if not SetFileTime(F.Handle, nil, nil, @ft) then
        raise EOSError.Create(F.Name);
    end;
    if Mode <> [] then
      if not (dmWriteByOwner in Mode) then
        if not SetFileAttributesW(PWideChar(F.Name),
            FILE_ATTRIBUTE_ARCHIVE or FILE_ATTRIBUTE_READONLY) then
          raise EOSError.Create(F.Name);
  end;
end;

procedure ClearDir(var Dir: TDir);
begin
  with Dir do begin
    Name := '';
    Size := 0;
    ATime := 0;
    MTime := 0;
    Mode := [];
  end;
end;

function SameFile(const Path1: WideString; const Path2: WideString): Boolean;
  function FileInfo(const Path: WideString): TByHandleFileInformation;
  var
    h: THandle;
  begin
    h:= CreateFileW(PWideChar(Path), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if h = INVALID_HANDLE_VALUE then
      raise EOSError.Create(Path);
    try
      if not GetFileInformationByHandle(h, Result) then
        raise EOSError.Create(Path);
    finally
      CloseHandle(h);
    end;
  end;
var
  info1, info2: TByHandleFileInformation;
begin
  info1 := FileInfo(Path1);
  info2 := FileInfo(Path2);
  Result := (info1.nFileIndexHigh = info2.nFileIndexHigh) and
      (info1.nFileIndexLow = info2.nFileIndexLow);
end;

function ReadLink(const Path: WideString): WideString;
var
  h: THandle;
  size: DWORD;
  buf: array [0..16383] of Char; // max reparse point data is 16 KiB
begin
  h := CreateFileW(PWideChar(Path), 0, FILE_SHARE_READ or FILE_SHARE_WRITE,
      nil, OPEN_EXISTING,
      FILE_FLAG_OPEN_REPARSE_POINT or FILE_FLAG_BACKUP_SEMANTICS, 0);
  if h = INVALID_HANDLE_VALUE then
    raise EOSError.Create(Path);

  try
    if not DeviceIoControl(h, FSCTL_GET_REPARSE_POINT, nil, 0, @buf,
        SizeOf(buf), size, nil) then
      raise EOSError.Create(Path);
  finally
    CloseHandle(h);
  end;

  with PREPARSE_DATA_BUFFER(@buf)^ do
    case ReparseTag of
      IO_REPARSE_TAG_MOUNT_POINT:
        with MountPointReparseBuffer do
          SetString(Result,
              PWideChar(@PathBuffer[SubstituteNameOffset div SizeOf(WCHAR)]),
              SubstituteNameLength div SizeOf(WCHAR));
      IO_REPARSE_TAG_SYMLINK:
        with SymbolicLinkReparseBuffer do
          SetString(Result,
              PWideChar(@PathBuffer[SubstituteNameOffset div SizeOf(WCHAR)]),
              SubstituteNameLength div SizeOf(WCHAR));
    else;
      Result := Path;
    end;
end;

function GetFileAccessRights(const Path: WideString): TFileAccessRights;
var
  i: Integer;
  dacl: PACL;
  asi: ACL_SIZE_INFORMATION;
  sd: PSECURITY_DESCRIPTOR;
  pace: PACCESS_ALLOWD_ACE;
  sdsize: DWORD;
  dp, dd: BOOL;
begin
  Result := [];
  sdsize := 65536; // max ACL size is 64 KiB
  // use LocalAlloc for alignment
  sd := Pointer(LocalAlloc(LMEM_FIXED, sdsize));
  try
    // may be faster than GetNamedSecurityInfoW on remote
    if not GetFileSecurityW(PWideChar(Path), DACL_SECURITY_INFORMATION, sd,
        sdsize, sdsize) then
      raise EOSError.Create;
    if not GetSecurityDescriptorDacl(sd, dp, dacl, dd) then
      raise EOSError.Create;
    if not dp then
      Exit;
    if dacl = nil then begin
      Include(Result, farRead);
      Include(Result, farWrite);
      Include(Result, farExecute);
      Exit;
    end;
    if not GetAclInformation(dacl^, @asi, sizeof(asi), AclSizeInformation) then
      raise EOSError.Create;

    for i := 0 to asi.AceCount-1 do begin
      if not GetAce(dacl^, i, Pointer(pace)) then
        raise EOSError.Create;
      if pace.Header.AceType <> ACCESS_ALLOWED_ACE_TYPE then
        Continue;
      if pace.Mask and FILE_READ_DATA <> 0 then Include(Result, farRead);
      if pace.Mask and FILE_WRITE_DATA <> 0 then Include(Result, farWrite);
      if pace.Mask and FILE_EXECUTE <> 0 then Include(Result, farExecute);
    end;
  finally
    LocalFree(HLOCAL(sd));
  end;
end;

function Time: Int64;
var
  ft: TFileTime;
begin
  GetSystemTimeAsFileTime(ft);
  Result := FileTimeToUnixTime(ft);
end;

function NanoSec: Int64;
var
  ft: TFileTime;
begin
  GetSystemTimeAsFileTime(ft);
  Result := (Int64(ft) - UnixTimeDiff) * 100;
end;

procedure SetNanoSec(const NanoSec: Int64);
var
  st: TSystemTime;
begin
  FileTimeToSystemTime(TFileTime(NanoSec div 100 + UnixTimeDiff), st);
  if not SetSystemTime(st) then
    raise EOSError.Create;
end;

function UTCTime(const T: Int64): TTime;
var
  st: TSystemTime;
begin
  FileTimeToSystemTime(UnixTimeToFileTime(T), st);
  with Result do begin
    Year := st.wYear;
    Mon := st.wMonth;
    MDay := st.wDay;
    Hour := st.wHour;
    Min := st.wMinute;
    Sec := st.wSecond;
    WDay := st.wDayOfWeek;
    UTCOffset := 0;
  end;
end;

function LocalTime(const T: Int64): TTime;
var
  ut, lt: TFileTime;
  st: TSystemTime;
begin
  ut := UnixTimeToFileTime(T);
  FileTimeToLocalFileTime(ut, lt);
  FileTimeToSystemTime(lt, st);
  with Result do begin
    Year := st.wYear;
    Mon := st.wMonth;
    MDay := st.wDay;
    Hour := st.wHour;
    Min := st.wMinute;
    Sec := st.wSecond;
    WDay := st.wDayOfWeek;
    UTCOffset := (Int64(lt) - Int64(ut)) div 10000000;
  end;
end;

procedure Sleep(MilliSecs: Cardinal);
begin
  Windows.Sleep(MilliSecs);
end;

procedure SplitDrive(const Path: WideString; var Drive, Tail: WideString);
begin
  if (Length(Path) > 1) and (Path[2] = DriveSep) then begin
    Drive := Copy(Path, 1, 2);
    Tail := Copy(Path, 3, Length(Path)-2);
  end else begin
    Drive := '';
    Tail := Path;
  end;
end;

procedure SplitExt(const Path: WideString; var Root, Ext: WideString);
var
  i: Integer;
  head, tail: WideString;
begin
  SplitPath(Path, head, tail);
  i := FindLastSep(tail, ExtSep);
  if i > 0 then begin
    Root := Copy(Path, 1, Length(Path)-(Length(tail)-i)-1);
    Ext := Copy(tail, i, Length(tail)-i+1);
  end else begin
    Root := Path;
    Ext := '';
  end;
end;

procedure SplitPath(const Path: WideString; var Head, Tail: WideString);
var
  i: Integer;
  p: WideString;
begin
  SplitDrive(Path, Head, p);
  i := FindLastPathSep(p);
  if i = 1 then
    Head := Head + p[1]
  else
    Head := Head + Copy(p, 1, i-1);
  Tail := Copy(p, i+1, Length(p)-i);
end;

function JoinPath(const Paths: array of WideString): WideString;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to High(Paths) do begin
    if Paths[i] = '' then
      Continue;
    if IsAbsPath(Paths[i]) then
      Result := '';
    if (FindLastPathSep(Paths[i]) = Length(Paths[i])) or (i = High(Paths)) then
      Result := Result + Paths[i]
    else
      Result := Result + Paths[i] + '\';
  end;
end;

function AbsPath(const Path: WideString): WideString;
var
  p: PWideChar;
  len: DWORD;
begin
  if Path = '' then begin
    Result := '';
    Exit;
  end;
  SetString(Result, nil, GetFullPathNameW(PWideChar(Path), 0, nil, p));
  len := GetFullPathNameW(PWideChar(Path), Length(Result), PWideChar(Result), p);
  if len = 0 then
    raise EOSError.Create;
  SetLength(Result, len);
end;

function IsAbsPath(const Path: WideString): Boolean;
var
  drive, tail: WideString;
begin
  SplitDrive(Path, drive, tail);
  Result := (tail <> '') and ((tail[1] = PathSep) or (tail[1] = AltPathSep));
end;

function StrColl(const S1, S2: string): Integer;
begin
  Result := lstrcmp(PChar(S1), PChar(S2));
end;

function StrCollW(const S1, S2: WideString): Integer;
begin
  Result := lstrcmpW(PWideChar(S1), PWideChar(S2));
end;

function StrUpper(const S: string): string;
begin
  SetString(Result, PChar(S), Length(S));
  CharUpperBuff(PChar(Result), Length(Result));
end;

function StrUpperW(const S: WideString): WideString;
begin
  SetString(Result, PWideChar(S), Length(S));
  CharUpperBuffW(PWideChar(Result), Length(Result));
end;

function DecodeUTF8(const S: string): WideString;
var
  len: Integer;
begin
  len := Length(S);
  SetString(Result, nil, len);
  SetLength(Result, MultiByteToWideChar(CP_UTF8, 0, PChar(S), len,
      PWideChar(Result), len));
end;

function EncodeUTF8(const S: WideString): string;
var
  len: Integer;
begin
  len := Length(S);
  SetLength(Result, len*4);
  SetLength(Result, WideCharToMultiByte(CP_UTF8, 0, PWideChar(S), len,
      PChar(Result), len*4, nil, nil));
end;

procedure Print(const S: string = ''; const End_: string = LineSep);
begin
  Print(StdOut, S, End_);
end;

procedure Print(const F: TStream; const S: string = '';
    const End_: string = LineSep);
begin
  F.Write((S + End_)[1], Length(S) + Length(End_));
end;

procedure Print(const S: string; const Args: array of const;
    const End_: string = LineSep);
begin
  Print(StdOut, S, Args, End_);
end;

procedure Print(const F: TStream; const S: string; const Args: array of const;
    const End_: string = LineSep);
begin
  Print(F, Format(S, Args), End_);
end;

function GetPass(const Prompt: string): string;
begin
  with TFile.OpenConIn do
    try
      Print(StdErr, Prompt, '');
      if IsEchoEnabled then
        try
          EnableEcho(False);
          ReadLine(Result);
        finally
          EnableEcho(True);
        end
      else
        ReadLine(Result);
    finally
      Free;
    end;
  RemoveLineSep(Result);
end;

function GenRandom(NBytes: Cardinal): string;
begin
  SetString(Result, nil, NBytes);
  if not RtlGenRandom(PChar(Result), NBytes) then
    raise EOSError.Create;
end;

{ TFile }

procedure TFile.CreateHandle(const Name: WideString; Mode: TFileMode;
  Create_: DWORD);
var
  access: DWORD;
begin
  FName := Name;
  access := 0;
  if fmRead in Mode then
    access := access or GENERIC_READ;
  if fmWrite in Mode then
    access := access or GENERIC_WRITE;
  FHandle := CreateFileW(PWideChar(FName), access,
    FILE_SHARE_READ or FILE_SHARE_WRITE, nil, Create_, FILE_ATTRIBUTE_NORMAL,
    0);
  if FHandle = INVALID_HANDLE_VALUE then
    if GetLastError = ERROR_FILE_NOT_FOUND then
      raise EFileNotFound.Create(FName)
    else
      raise EOSError.Create(FName);
end;

constructor TFile.Create(const Name: WideString; Mode: TFileMode=[fmWrite]);
begin
  CreateHandle(Name, Mode, CREATE_ALWAYS);
end;

constructor TFile.Open(const Name: WideString; Mode: TFileMode=[fmRead]);
begin
  CreateHandle(Name, Mode, OPEN_EXISTING);
end;

constructor TFile.Open(StdIO: TStdIO);
const
  names: array [0..2] of PWideChar = ('<STDIN>', '<STDOUT>', '<STDERR>');
begin
  FName := names[Ord(StdIO)];
  FHandle := GetStdHandle(DWORD(-10 - Ord(StdIO)));
  IsStdHandle := True;
end;

constructor TFile.OpenConOut;
begin
  Open('CONOUT$', [fmRead, fmWrite]);
end;

constructor TFile.OpenConIn;
begin
  Open('CONIN$', [fmRead, fmWrite]);
end;

constructor TFile.CreateByHandle(Handle: THandle);
begin
  FHandle := Handle;
end;

destructor TFile.Destroy;
begin
  if (FHandle <> INVALID_HANDLE_VALUE) and not IsStdHandle then
    CloseHandle(FHandle);
  inherited;
end;

function TFile.Size: Int64;
begin
  if not GetFileSizeEx(FHandle, LARGE_INTEGER(Result)) then
    raise EOSError.Create(FName);
end;

function TFile.Read(var Buf; NBytes: Cardinal; var NRead: Cardinal): LongBool;
begin
  if not ReadFile(FHandle, Buf, NBytes, NRead, nil) then
    if GetLastError <> ERROR_BROKEN_PIPE then // The pipe has been ended.
      raise EOSError.Create(FName);
  Result := LongBool(NRead);
end;

function TFile.Write(const Buf; NBytes: Cardinal): Cardinal;
begin
  if not WriteFile(FHandle, Buf, NBytes, Result, nil) then
    case GetLastError of
      ERROR_BROKEN_PIPE, // The pipe has been ended.
      ERROR_NO_DATA:     // The pipe is being closed.
        raise EPipeError.Create(FName);
    else
      raise EOSError.Create(FName);
    end;
end;

function TFile.Seek(Offset: Int64; Whence: TSeekWhence=swSet): Int64;
begin
  if not SetFilePointerEx(FHandle, LARGE_INTEGER(Offset),
      LARGE_INTEGER(Result), DWORD(Whence)) then
    raise EOSError.Create(FName);
end;

function TFile.Seekable: Boolean;
begin
  Result := GetFileType(FHandle) = FILE_TYPE_DISK;
end;

function TFile.IsConsole: Boolean;
begin
  Result := GetFileType(FHandle) = FILE_TYPE_CHAR;
end;

function TFile.IsEchoEnabled: Boolean;
var
  mode: DWORD;
begin
  if not GetConsoleMode(FHandle, mode) then
    raise EOSError.Create;
  Result := mode and ENABLE_ECHO_INPUT <> 0;
end;

procedure TFile.EnableEcho(Flag: Boolean);
var
  mode: DWORD;
begin
  if not GetConsoleMode(FHandle, mode) then
    raise EOSError.Create;
  if Flag then
    mode := mode or ENABLE_ECHO_INPUT
  else
    mode := mode and not ENABLE_ECHO_INPUT;
  if not SetConsoleMode(Handle, mode) then
    raise EOSError.Create;
end;

{ TListDir }

constructor TListDir.Create(const Path: WideString);
begin
  Self.Path := Path;
  Handle := FindFirstFileW(PWideChar(JoinPath([Self.Path, '*.*'])), FindData);
  if Handle = INVALID_HANDLE_VALUE then
    raise EOSError.Create(Self.Path);
  First := True;
end;

destructor TListDir.Destroy;
begin
  if Handle <> INVALID_HANDLE_VALUE then
    FindClose(Handle);
  inherited;
end;

function TListDir.Each(var Dir: TDir): Boolean;
begin
  if First then begin
    First := False;
    Dir := FindDataToDir(FindData);
    if (Dir.Name <> '.') and (Dir.Name <> '..') then begin
      Result := True;
      Exit;
    end;
  end;

  while FindNextFileW(Handle, FindData) do begin
    Dir := FindDataToDir(FindData);
    if (Dir.Name <> '.') and (Dir.Name <> '..') then begin
      Result := True;
      Exit;
    end;
  end;
  if GetLastError <> ERROR_NO_MORE_FILES then
    raise EOSError.Create(Path);

  Result := False;
end;

{ TProcess }

constructor TProcess.Create(const Args: array of WideString;
    const StdIn: TFile; const StdOut: TFile; const StdErr: TFile;
    ErrToOut: Boolean=False);
var
  si: TStartupInfo;
  inr, outw, errw: THandle;
  outwtmp, errwtmp: TFile;

  function dup(Handle: THandle): THandle;
  begin
    DuplicateHandle(GetCurrentProcess, Handle, GetCurrentProcess, @Result, 0,
        True, DUPLICATE_SAME_ACCESS);
    if Result = 0 then
      raise EOSError.Create;
  end;

begin
  FillChar(si, SizeOf(si), 0);
  si.cb := SizeOf(si);
  si.dwFlags := STARTF_USESTDHANDLES;
  inr := 0;
  outw := 0;
  errw := 0;

  try
    if StdIn.Handle = GetStdHandle(STD_INPUT_HANDLE) then
      si.hStdInput := StdIn.Handle
    else begin
      inr := dup(StdIn.Handle);
      si.hStdInput := inr;
    end;

    if StdOut = nil then begin
      CreatePipe(StdOutReader, outwtmp);
      outw := dup(outwtmp.Handle);
      outwtmp.Free;
      si.hStdOutput := outw;
    end else begin
      if StdOut.Handle = GetStdHandle(STD_OUTPUT_HANDLE) then
        si.hStdOutput := StdOut.Handle
      else begin
        outw := dup(StdOut.Handle);
        si.hStdOutput := outw;
      end;
    end;

    if ErrToOut then
      si.hStdError := si.hStdOutput
    else
      if StdErr = nil then begin
        CreatePipe(StdErrReader, errwtmp);
        errw := dup(errwtmp.Handle);
        errwtmp.Free;
        si.hStdError := errw;
      end else begin
        if StdErr.Handle = GetStdHandle(STD_ERROR_HANDLE) then
          si.hStdError := StdErr.Handle
        else begin
          errw := dup(StdErr.Handle);
          si.hStdError := errw;
        end;
      end;

    if not CreateProcessW(nil, PWideChar(JoinCommandLine(Args)), nil, nil,
        True, 0, nil, nil, si, ProcessInformation) then
      raise EOSError.Create(Args[0]);
  finally
    if inr <> 0 then CloseHandle(inr);
    if outw <> 0 then CloseHandle(outw);
    if errw <> 0 then CloseHandle(errw);
  end;
end;

destructor TProcess.Destroy;
begin
  CloseHandle(ProcessInformation.hProcess);
  CloseHandle(ProcessInformation.hThread);
  StdOutReader.Free;
  StdErrReader.Free;
end;

function TProcess.Wait;
begin
  WaitForSingleObject(ProcessInformation.hProcess, INFINITE);
  GetExitCodeProcess(ProcessInformation.hProcess, DWORD(Result));
end;

procedure TProcess.Times(var System, User: Int64);
var
  ct, et, kt, ut: FILETIME;
begin
  if not GetProcessTimes(ProcessInformation.hProcess, ct, et, kt, ut) then
    raise EOSError.Create;
  System := Int64(kt) div 10000;
  User := Int64(ut) div 10000;
end;

initialization
  Argv := SplitCommandLineW(GetCommandLineW);
  ProgName := PWideChar(@Argv[0][FindLastPathSep(Argv[0]) + 1]);
end.
